;################################################
; perform RM reconstruction on given subset of data
;################################################
; Input:
; => phi_maps     : stack pol direction maps
; => phia_maps    : stack flags for pixels with abs phi defined
; => err_maps     : stac of pol direction error maps
; => mask         : mask giving the valid region
; => frqs         : array of the according frequencies
; => rm_all       : map containing global RM map
; => patch_map    : map containing the patch numbers
; Output:
; <= rmout        : map containing the RM values
; <= iphi         : map of phi_0 in the source
; <= xi2map       : map with the xi2 values of the fit
; <= rejects      : map for rejected pixel
; <= rm0map       : map with the standard RM fit values
; <= phi0map      : map of phi_0 in the source for the standard RM fit
; <= flags1       : map with the flags of pixel
; <= squality     : map containing the quality of pixels
; Optional:
; => start_qual   : quality of pixel to be starting point
; => show         : show graphically progress in algorithm
; => zoom         : resize of images to show
; => noweight     : supress weighting by errors for RM fit
; => warning      : enables different debug output
; => grad_tol     : sets quality gradient constrains
; => show         : flag to display progress
; => rmmax        : set maximum of RM expected
; => rmmin        : set minimum of RM expected
; => save         : saves output of show into a file, to produce movies
; => max_dphi_toll: maximum of tolerated delta phi betwen pixels
; => democaratic  : allows democratic grocess for solving global n-Pi

PRO make_smooth,phi_maps,phia_maps,err_maps,mask,frqs,rm_all,squality,$
                rmout,iphi,xi2map,rejects,$
                rm0map,phi0map,flags1,patch_map,$
                start_qual=start_qual,do_rm0=do_rm0,max_dphi_toll=max_dphi_toll,$
                show=show,zoom=zoom,noweight=noweight,democratic=democratic,$
                warning=warning,grad_tol=grad_tol,rmmax=rmmax,save=save,rmmin=rmmin,$
                aips=aips,xi2_frac_cut=xi2_frac_cut,min_neighb=min_neighb
common patch,patchnr
   IF NOT(IS_DEF(zoom)) THEN zoom=1
   IF NOT(IS_DEF(warning)) THEN warning=0
   IF NOT(IS_DEF(grad_tol)) THEN grad_tol=1.2
   IF NOT(IS_DEF(start_qual)) THEN start_qual=MAX(squality)
   IF NOT(IS_DEF(xi2_frac_cut)) THEN xi2_frac_cut=0.75
;  phi_maps in [degrees]
;  frqs in [MHz]
;                           Define neighbours
   idx =[-1, 0, 1,-1, 1,-1, 0, 1]
   idy =[ 1, 1, 1, 0, 0,-1,-1,-1]
   wght=[1.,1.,1.,1.,1.,1.,1.,1.]
   anz=8
   ishow=0
;  weight parameter:
;     0.0 => Dont benefit fron neighbours
;     1.0 => Benefit from neightbours due to geometrical mean
   alpha=0.5
   alpha_ini=0.1
;                           Test size of maps
   packmap=[[[0,0,1,1,0,0],$
             [0,1,1,1,1,0],$
             [1,1,1,1,1,1],$
             [1,1,1,1,1,1],$
             [0,1,1,1,1,0],$
             [0,0,1,1,0,0]],$
            [[0,0,1,1,0,0],$
             [0,1,1,1,1,1],$
             [1,1,1,1,1,0],$
             [1,1,1,1,1,0],$
             [0,1,1,1,1,1],$
             [0,0,1,1,0,0]],$
            [[0,0,1,1,0,0],$
             [0,1,1,1,1,1],$
             [1,1,1,1,0,0],$
             [1,1,1,1,0,0],$
             [0,1,1,1,1,1],$
             [0,0,1,1,0,0]],$
            [[0,0,1,1,0,1],$
             [0,1,1,1,1,0],$
             [1,1,1,1,0,0],$
             [1,1,1,1,0,0],$
             [0,1,1,1,1,0],$
             [0,0,1,1,0,1]]]


;
   ss=SIZE(phi_maps)
   nmaps=ss(1)
   xsize=ss(2)
   ysize=ss(3)
;                           Define working maps
   flags1=fltarr(xsize,ysize)
;   patch_map=fltarr(xsize,ysize)
;
   done=fltarr(xsize,ysize)
   todo=fltarr(xsize,ysize)
   doable=fltarr(xsize,ysize)
   rejects=fltarr(xsize,ysize)
   xi2map=fltarr(xsize,ysize)
   ixmap=intarr(xsize,ysize)
   iymap=intarr(xsize,ysize)
   phi=fltarr(nmaps,xsize,ysize)
   phia=fltarr(nmaps,xsize,ysize)
   rmout=fltarr(xsize,ysize)
   iphi=fltarr(xsize,ysize)
   rm0map=fltarr(xsize,ysize)
   phi0map=fltarr(xsize,ysize)
   nphilist=fltarr(nmaps,anz)
;                           Fill working maps
   todo(*,*)=0.0
   doable(*,*)=0.0
   done(*,*)=0.0
   rejects(*,*)=0.0
   xi2map(*,*)=-1.0
   phi(*,*,*)=phi_maps(*,*,*)
   phia(*,*,*)=phia_maps(*,*,*)

   lamb2=(2.99792458d+8/(frqs*1.0d6))^2

   FOR ix=0,xsize-1 DO BEGIN
      FOR iy=0,ysize-1 DO BEGIN
         ixmap(ix,iy)=ix
         iymap(ix,iy)=iy
         IF (ix GT 0) AND (iy GT 0) AND (ix LT xsize-1) AND (iy LT ysize-1) THEN BEGIN
; Improve quality by presents of neighbours from previous fitted patches ...
            todo(ix,iy)=0.0
            numneigh=0
            FOR iqn=0L,anz-1 DO BEGIN
               IF((rm_all(ix+idx(iqn),iy+idy(iqn)) NE 0.0) AND $
                  (MIN(phia_maps(*,ix+idx(iqn),iy+idy(iqn))) GT 0.0)) THEN BEGIN
; fill list to test if n-Pi transport is possible
                   nphilist(*,numneigh)=phi(*,ix+idx(iqn),iy+idy(iqn))
                   numneigh=numneigh+1
               END
            END
            IF numneigh GT 0 THEN BEGIN
; test n-Pi from neighbours found ...
               phi_v=phi(*,ix,iy)
               find_near_abs_phi,phi_v,nphilist,numneigh,maskfl,warning=warning,max_dphi_toll=max_dphi_toll,itmp_use=itmp_use
               IF (maskfl LE 1.0) THEN BEGIN
                  IF squality(ix,iy) LE start_qual THEN BEGIN
                     todo(ix,iy)=squality(ix,iy)*(0.97)^itmp_use
                  END ELSE BEGIN
                     todo(ix,iy)=start_qual*((alog10(squality(ix,iy)+start_qual)-alog10(start_qual))/alog10(squality(ix,iy)+start_qual))^itmp_use
                  END
;                  todo(ix,iy)=MIN([start_qual/2.,squality(ix,iy)/2])
               END ELSE BEGIN
                  todo(ix,iy)=squality(ix,iy)
               END
            END ELSE BEGIN
               todo(ix,iy)=squality(ix,iy)
            END
         END ELSE BEGIN
            todo(ix,iy)=squality(ix,iy)
         END
      END
   END

;                          Start with full mask
   tmpmask=mask
   jjmask=where(tmpmask GT 0.0)

   count_out=0L
   print_out=100L

   ss=SIZE(jjmask)

   phi_in=fltarr(nmaps)
   err_in=fltarr(nmaps)

   IF(IS_DEF(do_rm0)) THEN BEGIN
      print,'Doing normal fit ... (RM_max=',rmmax,')'
; avoid to fit new data on existing points ...
      jj0mask=where((tmpmask GT 0.0) AND (rm0map EQ 0.0))
iiis=SORT(todo(jj0mask))
      ss0=SIZE(jj0mask)
      FOR i=0L,ss0(1)-1 DO BEGIN
         phi_in(*)=phi_maps(*,ixmap(jj0mask(iiis(i))),iymap(jj0mask(iiis(i))))
         err_in(*)=err_maps(*,ixmap(jj0mask(iiis(i))),iymap(jj0mask(iiis(i))))

         fit_rm,lamb2,phi_in,err_in,phi_ini,rm,xi2,flags,rmmax=rmmax,rmmin=rmmin,$
                noweight=noweight,warning=warning,aips=aips

         rm0map(ixmap(jj0mask(iiis(i))),iymap(jj0mask(iiis(i))))=rm
         phi0map(ixmap(jj0mask(iiis(i))),iymap(jj0mask(iiis(i))))=phi_ini
;         flags1(ixmap(jj0mask(iiis(i))),iymap(jj0mask(iiis(i))))=flags
         count_out=count_out+1
         IF count_out GE print_out THEN BEGIN
            print,'Fitted ',i,' of ',ss0(1),' Points ...'
            count_out=0L
         END
      END
   END

   IF IS_DEF(show) THEN erase,0
;                          Loop over all paches
   count_out=0L
   fcount=0L
   WHILE jjmask(0) GE 0 DO BEGIN
;                          find first Point
      minval=MIN(todo(jjmask))
      jjs=where(todo(jjmask) EQ minval)
      ixs=ixmap(jjmask(jjs(0)))
      iys=iymap(jjmask(jjs(0)))
IF minval GT start_qual THEN BEGIN
         sstmp=SIZE(jjmask)
         print,'   Rejected Starting Point:  ',ixs,iys,minval,squality(ixs,iys),start_qual
         print,'   (',sstmp(1),' Points skiped)'
         tmpmask(jjmask)=0.0
END ELSE BEGIN
      print,'Doing Patch',patchnr,' ...'
      print,'   Starting Point: ',ixs,iys,minval,start_qual
      doable(ixs,iys)=1.0
;                          DEFINE working area
      jjdo=WHERE(doable GT 0.0)
      count=0L
      no_n_pi=0
      guidepatch=0
      startreject=0
;                          Go through conected area
      WHILE jjdo(0) GE 0 DO BEGIN
;                          find next point
         minval=MIN(todo(jjdo))
         jj=where(todo(jjdo) EQ minval)
         ix=ixmap(jjdo(jj(0)))
         iy=iymap(jjdo(jj(0)))
;                          Select neightbours
         itmp=0
         maskfl=1.0
;
         order=SORT(ABS(patch_map(ix+idx(*),iy+idy(*))))
         FOR in_tmp=0L,anz-1 DO BEGIN
            in=order(in_tmp)
            itx=ix+idx(in)
            ity=iy+idy(in)
            IF done(itx,ity) GT 0.0 THEN BEGIN
               nphilist(*,itmp)=phi(*,itx,ity)
               itmp=itmp+1
            END ELSE BEGIN
; transport informations across multible frequencie maps
               IF (rm_all(itx,ity) NE 0.0) AND (maskfl LE 1.0) AND $
                  (MIN(phia_maps(*,itx,ity)) GT 0.0) THEN BEGIN
; global n-Pi, only for first point, as the new patch might have different
; (already solved) patches as neighbour.
                  IF (count EQ 0) AND (no_n_pi EQ 0) THEN BEGIN
                     print,'Transported n-Pi information ...'
                     nphi0merk=nphilist(*,0)
                     itmp_merk=itmp
                     nphilist(*,0)=phi(*,itx,ity)
                     itmp=1
                     phi_v=phi(*,ix,iy)
                     find_near_abs_phi,phi_v,nphilist,itmp,maskfl,warning=warning,max_dphi_toll=max_dphi_toll
                     IF (maskfl LE 1.0) THEN no_n_pi=1
                     IF no_n_pi EQ 1 THEN BEGIN
                        ixsm=itx
                        iysm=ity
                        guidepatch=abs(patch_map(itx,ity))
                        startreject=0
                     END ELSE BEGIN
                        print,'Rejected ...'
; stop this patch
;                        itmp=0
; go on with this patch
                        itmp=itmp_merk
                        nphilist(*,0)=nphi0merk(*,*)
                        maskfl=1
                        startreject=1
                     END
                  END ELSE BEGIN
; Also use only one neighbour from one other patch ...
                     IF guidepatch EQ patch_map(itx,ity) THEN BEGIN
;                        print,'Transport additional neighbour ...'
                        nphilist(*,itmp)=phi(*,itx,ity)
                        itmp=itmp+1
                     END
                  END
               END
            END
         END
;IF patchnr EQ 184 THEN stop
         IF ((no_n_pi EQ 0) OR (count GT 0)) AND (maskfl LE 1.0) THEN BEGIN
;                          Calculate absulute angle using the
;                          informations provided by the neightbours
            phi_v=phi(*,ix,iy)
            find_near_abs_phi,phi_v,nphilist,itmp,maskfl,warning=warning,max_dphi_toll=max_dphi_toll
; should not happen, check data for NaN's ...
            IF (MAX(ABS(phi_v)) GT 2000) THEN stop
         END ELSE BEGIN
            IF (no_n_pi EQ 1) AND (count EQ 0) AND (guidepatch GT 0) THEN BEGIN
               find_near_abs_phi,phi_v,nphilist,itmp,maskfl,warning=warning,max_dphi_toll=max_dphi_toll,itmp_use=itmp_use
               IF (((itmp_use LT itmp/2) OR (itmp LE min_neighb)) AND (min_neighb GT 0)) $
                    OR ((min_neighb EQ 0) AND (itmp_use EQ 0)) THEN BEGIN
                  maskfl=2
                  startreject=1
                  guidepatch=0
                  no_n_pi=0
               END
            END 
         END
;                          Test if this point looks not quite clear
         IF (maskfl GT 1.0) THEN BEGIN
;                          Put it back in que
               IF (warning GE 3) THEN print,'Putting point back in que ...'
               IF todo(ix,iy) LT squality(ix,iy) THEN todo(ix,iy)=squality(ix,iy) ELSE todo(ix,iy)=todo(ix,iy)*2
               rejects(ix,iy)=rejects(ix,iy)+1
               IF (rejects(ix,iy) GT 5) THEN BEGIN
                  IF (warning GE 1) THEN print,'WARNING: Rejecting obscure data Point ...'
                  doable(ix,iy)=0.0
                  tmpmask(ix,iy)=0.0       ; never try thins point again !
               END
         END ELSE BEGIN
;IF patchnr EQ 3508 THEN stop
            phi(*,ix,iy)=phi_v(*)
            done(ix,iy)=1.0
;                          Go through neighbours and find pixels
;                          which could be done next
            FOR in=0L,anz-1 DO BEGIN
               itx=ix+idx(in)
               ity=iy+idy(in)
               IF tmpmask(itx,ity) GT 0.0 THEN BEGIN
;                          Inprove quality by neighbours with defined abs phi
                 numneigh=0
;                 todo(itx,ity)=0.0
                 FOR iqn=0L,anz-1 DO BEGIN
                    IF(done(itx+idx(iqn),ity+idy(iqn)) EQ 1) THEN BEGIN
;                       todo(itx,ity)=todo(itx,ity)+wght(iqn)*1./squality(itx+idx(iqn),ity+idy(iqn))
                       numneigh=numneigh+1
                    END
                 END
                 IF numneigh GT 0 THEN todo(itx,ity)=0.975^numneigh*squality(itx,ity)
;                 IF numneigh GT 0 THEN todo(itx,ity)=alpha/numneigh*todo(itx,ity)
;                 todo(itx,ity)=todo(itx,ity)+1./squality(itx,ity)

;                          Only go toward less defined points
                 IF todo(ix,iy) LT grad_tol*squality(itx,ity) THEN BEGIN
                     doable(itx,ity)=1.0
                 END ELSE BEGIN
;                    IF warning GE 1 THEN
                        print,'Rejected to go from',todo(ix,iy),' -> ',squality(itx,ity)
                 END
               END
            END
;                          Remove done Pixel from todo
            doable(ix,iy)=0.0
;                          Remove Pixel from overall pixels to do
            tmpmask(ix,iy)=0.0
         END
;
;                          DEFINE working area
         jjdo=WHERE(doable GT 0.0)
         IF (maskfl LE 1.0) THEN count=count+1L

         count_out=count_out+1
         IF count_out GE print_out THEN BEGIN
      jjdone=WHERE(done GE 1.0)
      ssdone=SIZE(jjdone)
            print,'Assigned ',count,' of ',ss(1),' Points ...',ssdone(1)
            count_out=0L
         END
      END
;                          Now fit this region
;                          How many points to fit
      jjdone=WHERE(done GE 1.0)

      IF (jjdone(0) GT -1) THEN BEGIN
         ssdone=SIZE(jjdone)
         IF (ssdone(1) LE 1) AND (startreject EQ 1) THEN BEGIN
            print,'WARNING: Remove single, obscure Point from list ...'
            doable(jjdone)=0.0
            tmpmask(jjdone)=0.0
            done(jjdone)=0.0
         END
      END

      jjdone=WHERE(done GE 1.0)

      IF jjdone(0) GT -1 THEN BEGIN
         ssdone=SIZE(jjdone)
         phi_vec=fltarr(nmaps,ssdone(1))
         err_vec=fltarr(nmaps,ssdone(1))
;                          Make vector with only datapoints
         istart=-1
         isort=SORT(todo(jjdone))
         FOR i=0L,ssdone(1)-1 DO BEGIN
            ix=ixmap(jjdone(isort(i)))
            iy=iymap(jjdone(isort(i)))
            IF i EQ 0 THEN BEGIN
               print,' => ',ix,iy
;               IF no_n_pi EQ 0 THEN BEGIN
                  ixsm=ix
                  iysm=iy
;               END
            END
            FOR n=0,nmaps-1 DO BEGIN
               phi_vec(n,i)=phi(n,ix,iy)
               err_vec(n,i)=err_maps(n,ix,iy)
            END
         END
         print,'Fitting patch with ',ssdone(1),' pixels ...'
         fit_rm,lamb2,phi_vec,err_vec,phi_ini,rm_vec,xi2map_vec,flags,rmmin=rmmin,$
                rmmax=rmmax,noweight=noweight,warning=warning,no_n_pi=no_n_pi,$
                democratic=democratic,aips=aips
         xi2=TOTAL(xi2map_vec)/ssdone(1)
;                          Sort in RM values
;IF patchnr EQ 205 THEN stop
         FOR i=0L,ssdone(1)-1 DO BEGIN
          ix=ixmap(jjdone(isort(i)))
          iy=iymap(jjdone(isort(i)))
          accept=1
          IF NOT(IS_DEF(democratic)) THEN BEGIN
             IF (flags(0) GT xi2_frac_cut) THEN accept=0
          END ELSE BEGIN
             IF (ssdone(1) GE democratic) AND (flags(0) GT xi2_frac_cut) THEN accept=0
             IF (ssdone(1) LT democratic) AND (flags(0) GT xi2_frac_cut*ssdone(1)/FLOAT(democratic)) THEN accept=0
          END
          IF accept EQ 0 THEN BEGIN
             IF i EQ 0 THEN print,'Rejecting patch due to marginal best fit n-pi ...',flags(0),xi2_frac_cut
          END ELSE BEGIN
            rmout(ix,iy)=rm_vec(i)
            IF ABS(rm_vec(i)) GT rmmax THEN print,'WARNING: Very big RM in point ',ix,iy,rm_vec(i)
            iphi(ix,iy)=phi_ini(i)
            xi2map(ix,iy)=xi2map_vec(i)
            flags1(ix,iy)=flags(i)
            FOR n=0,nmaps-1 DO BEGIN
               phi(n,ix,iy)=phi_vec(n,i)
               phia(n,ix,iy)=1.
            END
            patch_map(ix,iy)=patchnr
          END
;                          Unmark the old area
          done(ix,iy)=0.0
         END
         IF (flags(0) LE xi2_frac_cut) THEN patch_map(ixsm,iysm)=-1*patchnr

      END
      patchnr=patchnr+1
 END
;                          DEFINE new mask of remaining area
      jjmask=where(tmpmask GT 0.0)
   END
   phi_maps(*,*,*)=phi(*,*,*)
   phia_maps(*,*,*)=phia(*,*,*)
   print,'Splitted map into ',patchnr,' Patches'
end
