; #################################
; Pacerman allgorithm for RM fitting
; #################################
; Needs:
; is_def.pro
; fit_rm.pro
; find_near_abs_phi.pro
; find_topological_defects.pro
; get_dataset.pro
; make_smooth.pro
; make_mask.pro

;################################################
;################################################
; Produce a RM MAP !
;################################################
;################################################
; Call
; IDL> pacerman,"data-identifier","pacerman_prefix"
;
; Input:
; => "data-identifier": Name of Source (identical with one of the blocks in get_dataset.pro)
; => "pacerman_prefix" :  Prefix for all output files produced (include here 
; 			  also the path for your output files)
;
; Optional Parameters:
; democratic=n          ; Uses the n best constrained point to infere'
;                       ; n-pi for a given patch (default=10???)
; rmmax=val             ; val = maximal allowed RM (used when solving n-pi)
; rmmin=val             ; val = minimal allowed RM (used when solving n-pi)
; /aips                 ; if defined, use first two frequencies to solve
;                       ; n-pi, instead of using rmmin and rmmax
; /do_rm0               ; additionally perform a standard RM fit to full map
; nmin_frq=val          ; val = number of frequencies needed to perform RM fit.
;                       ;       (can be snmaller than the total number of
;                       ;        frequencies observed, default =
;                       ;        number of frequencies)
; grad_tol=val          ; val = Maximal gradient allowed for going from low to high
;                       ;       quality data, default: 1.2
; max_dphi_toll=val     ; val = maximum phi variation between neighbouring pixels
;                       ;       within one patch default 15
; phi_err_tre           ; val = maximum allowed error in phi for a pixel
;                       ;       to be a valid pixel default 15
; /noweight             ; turn off the weighting by the errors for the RM fit
;
; Output:
; <prefix>.fits         ; final RM Map
; <prefix>_flags1.fits  ; flags for pixels (used to look why pixels got excluded)
; <prefix>_nfrq.fits    ; Number of valid frequencies for each pixel
; <prefix>_patch.fits   ; Number of patch to which each pixel belongs
; <prefix>_phi<n+1>.fits; maps of the reconstructed absolute value of
;                       ; phi. phi0 is the reconstructed intrinsic orientation
; <prefix>_quality.fits ; gives the data quality of eaxh pixel
; <prefix>_rej.fits     ; gives how often each pixel was rejected
;                       ; during the fitting procedure.
; <prefix>_rm1.fits     ; contains the standart fit (if /do_rm0) was given
; <prefix>_rm1_phi0.fits; reconstructed internal orientation for standart fit
; <prefix>_xi2.fits     ; Xi^2 for the RM fit


pro pacerman, source,save,democratic=democratic,rmmax=rmmax,rmmin=rmmin,aips=aips,$
            do_rm0=do_rm0,nmin_frq=nmin_frq,grad_tol=grad_tol,max_dphi_toll=max_dphi_toll,$
            phi_err_tre=phi_err_tre,noweight=noweight,subset=subset,$
; Following parameters are none standart parameters used mainly for testing
            rmmap=rmmap,rmomap=rmomap,iphimap=iphimap,errmap=errmap,$
            show=show,zoom=zoom,sto=sto,afterburn=afterburn,min_neighb=min_neighb,$
            warning=warning,clean=clean,start_tresh=start_tresh,xi2_frac_cut=xi2_frac_cut
;            start_abs=start_abs
common patch,patchnr

;   setcolors

   IF NOT(IS_DEF(source)) OR NOT(IS_DEF(save)) THEN BEGIN
      print,'You have to give at least a dataset-id and a name to save !'
      print,'Syntax:'
      print,'IDL> pacerman,"HydraN","pacerman_prefix_"'
      print,'optional:'
      print,'   democratic=n          ; Uses the n best constrained point to infere'
      print,'                         ; n-pi for a given patch'
      print,'   rmmax=val             ; val = maximal allowed RM (used when solving n-pi)'
      print,'   rmmin=val             ; val = minimal allowed RM (used when solving n-pi)'
      print,'   /aips                 ; if defined, use first two frequencies to solve'
      print,'                         ; n-pi, instead of using rmmin/rmmax'
      print,'   /do_rm0               ; additionally perform a standard RM fit to full map'
      print,'   nmin_frq_tol=n        ; n = number of frequencies needed to perform RM fit.'
      print,'                         ;       (can be snmaller than the total number of '
      print,'                         ;        frequencies observed)'
      print,'   grad_tol=val          ; val = Maximal gradien allowed for going from low to'
      print,'                         ;       high quality data, default: 1.2'
      print,'   max_dphi_tol=val      ; val = maximum phi variation between neighbouring'
      print,'                         ;       pixels within one patch (default 15)'
      print,'   phi_err_tre           ; val = maximum allowed error in phi for a pixel'
      print,'                         ;       to be a valid pixel (default 15)'
      print,'   start_tresh           ; val = maximum allowed quality for a pixel'
      print,'                         ;       to be a valid patch start pixel 
                                      ;       (default 2)'
      print,'   /noweight             ; turn off the weighting by the errors for the RM fit'
      return
   END

;################################################
; define the location and names of maps available
;################################################
   get_dataset,source,found,frq,phi_pol,phi_pol_err,group,$
                ix0min=ix0min,ix0max=ix0max,iy0min=iy0min,iy0max=iy0max,path=path,$
                phi_err_tre=phi_err_tre,frq_need=frq_need,subset=subset,err_mul=err_mul

   ss=SIZE(frq)
   nfr=ss(1)

   IF NOT(IS_DEF(rmmax)) THEN rmmax=5000.0                            ; max. expected RM
   IF NOT(IS_DEF(grad_tol)) THEN grad_tol=1.1                         ; Toteranze for quality gradient
   IF NOT(IS_DEF(nmin_frq)) THEN nmin_frq=nfr                         ; min number of frq to fit
   IF NOT(IS_DEF(start_tresh)) THEN start_tresh=2.0
   IF NOT(IS_DEF(max_dphi_toll)) THEN max_dphi_toll=15
   IF NOT(IS_DEF(phi_err_tre)) THEN phi_err_tre=15.0              ; aceptance of err
   IF NOT(IS_DEF(min_neighb)) THEN min_neighb=1

   start_tresh_tmp=fltarr(nfr-nmin_frq+1)
   start_tresh_tmp(*)=-1
   ss=SIZE(start_tresh)
   IF ss(0) EQ 0 THEN start_tresh_tmp(*)=start_tresh
   IF ss(0) EQ 1 AND ss(1) EQ nfr-nmin_frq+1 THEN start_tresh_tmp(*)=start_tresh(*)
   IF start_tresh_tmp(0) EQ -1 THEN BEGIN
      print,'start_tresh has not the right dimension !',ss
      print,'You can:'
      print,'1) use the default by not specifying <start_tresh> in the call'
      print,'2) give a single value which is used for all combinations of frequencies'
      print,'3) give an array of size (number of given frequency maps - nmin_frq + 1)'
   END ELSE start_tresh=start_tresh_tmp

   max_dphi_toll_tmp=fltarr(nfr-nmin_frq+1)
   max_dphi_toll_tmp(*)=-1
   ss=SIZE(max_dphi_toll)
   IF ss(0) EQ 0 THEN max_dphi_toll_tmp(*)=max_dphi_toll
   IF ss(0) EQ 1 AND ss(1) EQ nfr-nmin_frq+1 THEN max_dphi_toll_tmp(*)=max_dphi_toll(*)
   IF max_dphi_toll_tmp(0) EQ -1 THEN BEGIN
      print,'max_dphi_toll has not the right dimension !',ss
      print,'You can:'
      print,'1) use the default by not specifying <max_dphi_toll> in the call'
      print,'2) give a single value which is used for all combinations of frequencies'
      print,'3) give an array of size (number of given frequency maps - nmin_frq + 1)'
   END ELSE max_dphi_toll=max_dphi_toll_tmp

   min_neighb_tmp=fltarr(nfr-nmin_frq+1)
   min_neighb_tmp(*)=-1
   ss=SIZE(min_neighb)
   IF ss(0) EQ 0 THEN min_neighb_tmp(*)=min_neighb
   IF ss(0) EQ 1 AND ss(1) EQ nfr-nmin_frq+1 THEN min_neighb_tmp(*)=min_neighb(*)
   IF max_dphi_toll_tmp(0) EQ -1 THEN BEGIN
      print,'min_neighb has not the right dimension !',ss
      print,'You can:'
      print,'1) use the default by not specifying <min_neighb> in the call'
      print,'2) give a single value which is used for all combinations of frequencies'
      print,'3) give an array of size (number of given frequency maps - nmin_frq + 1)'
   END ELSE min_neighb=min_neighb_tmp

   phi_err_tre_tmp=fltarr(nfr)
   phi_err_tre_tmp(*)=-1
   ss=SIZE(phi_err_tre)
   IF ss(0) EQ 0 THEN phi_err_tre_tmp(*)=phi_err_tre
   IF ss(0) EQ 1 AND ss(1) EQ nfr THEN phi_err_tre_tmp(*)=phi_err_tre(*)
   IF phi_err_tre_tmp(0) EQ -1 THEN BEGIN
      print,'phi_err_tre has not the right dimension !',ss
      print,'You can:'
      print,'1) use the default by not specifying <phi_err_tre> in the call'
      print,'2) give a single value which is used for all frequencies'
      print,'3) give an array of size (number of given frequency maps)'
   END ELSE phi_err_tre=phi_err_tre_tmp

   IF NOT(IS_DEF(frq_need)) THEN frq_need=0
   frq_need_tmp=fltarr(nfr)
   frq_need_tmp(*)=-1
   ss=SIZE(frq_need)
   IF ss(0) EQ 0 THEN frq_need_tmp(*)=frq_need
   IF ss(0) EQ 1 AND ss(1) EQ nfr THEN frq_need_tmp(*)=frq_need(*)
   IF frq_need_tmp(0) EQ -1 THEN BEGIN
      print,'frq_need has not the right dimension !',ss
      print,'You can:'
      print,'1) use the default(=0) by not specifying <frq_need> in get_dataset'
      print,'2) give a single value (0 or 1) which is used for all frequencies'
      print,'3) give an array of size (number of given frequency maps)'
   END ELSE frq_need=frq_need_tmp

   IF IS_DEF(afterburn) THEN phi_err_tre_after=phi_err_tre*ABS(afterburn)

   frq_need_sum=TOTAL(frq_need)

;################################################
; Load data and constructing a proper mask
;################################################
   make_mask,frq,map_stack,err_stack,mask_stack,mask,$
              phi_pol_file=phi_pol,phi_pol_err=phi_pol_err,$
              path=path,show=show,$
              ix0min=ix0min,ix0max=ix0max,iy0min=iy0min,iy0max=iy0max,nmin_frq=nmin_frq,$
              phi_err_tre=phi_err_tre,header=header,err_mul=err_mul

   sshead=SIZE(header)
   FOR ih=0,sshead(1)-1 DO BEGIN
      hstr=header(ih)
      t=STRPOS(hstr,'BUNIT')
      IF t EQ 0 THEN BEGIN
         print,'replacing BUNIT keyword ....'
         header(ih)="BUNIT   = 'RAD/M/M '           /UNITS OF FLUX"
      END
   END
;stop
;#######################################################
; reducing to needed size ...
;#######################################################
   print,'Reducing size ...'
   ss=SIZE(mask)
   minix=ss(1)-1
   maxix=0
   miniy=ss(2)-1
   maxiy=0
   FOR ix=0,ss(1)-1 DO BEGIN
      FOR iy=0,ss(2)-1 DO BEGIN
         IF (mask(ix,iy) GT 0.0) THEN BEGIN
            IF (ix LT minix) THEN minix=ix
            IF (iy LT miniy) THEN miniy=iy
            IF (ix GT maxix) THEN maxix=ix
            IF (iy GT maxiy) THEN maxiy=iy
         END
      END
   END
   print,'=> ',minix,maxix,miniy,maxiy
;   minix=MAX([minix-2,0])
;   maxix=MIN([maxix+2,ss(1)-1])
;   miniy=MAX([miniy-2,0])
;   maxiy=MIN([maxiy+2,ss(2)-1])

   dx=maxix-minix+1
   dy=maxiy-miniy+1

   rmout=fltarr(dx,dy)                    ; Final RM map
   flags1=fltarr(dx,dy)                   ; Flags for First guess RM map
   patch_map=fltarr(dx,dy)                ; Map for Patch number
   rm1map=fltarr(dx,dy)                   ; First guess RM map
   phi1map=fltarr(dx,dy)                  ; First guess RM map

   iphi=fltarr(dx,dy)                     ; polarisation direction in source
   smap_stack=fltarr(nfr,dx,dy)           ; polarisation direction at the frqs
   serr_stack=fltarr(nfr,dx,dy)           ; errors of the pol. directions

   smask_stack=fltarr(nfr,dx,dy)          ; mask for pol>pol_tresh
   smask=fltarr(dx,dy)                    ; mask for nfrq(pol>pol_tresh)>min_frq

   sphia_stack=fltarr(nfr,dx,dy)          ; mask for pol>pol_tresh

   nrmask=fltarr(dx,dy)                   ; freq. used for RM fit
   xi2map=fltarr(dx,dy)                   ; Xi2 map
   squality=fltarr(dx,dy)                 ; over all quality of RM fit

   smask_part=fltarr(dx,dy)               ; final mask for actual fit
   smask_tmp=fltarr(dx,dy)                ; temp mask


   nrmask(*,*)=0.0
   sphia_stack(*,*,*)=0.0
   smask(*,*)=mask(minix:maxix,miniy:maxiy)
   smask(0,*)=0 & smask(dx-1,*)=0 & smask(*,0)=0 & smask(*,dy-1)=0
   smap_stack(*,*,*)=map_stack(*,minix:maxix,miniy:maxiy)
   smask_stack(*,*,*)=mask_stack(*,minix:maxix,miniy:maxiy)
   serr_stack(*,*,*)=err_stack(*,minix:maxix,miniy:maxiy)
   squality(*,*)=0.0


;#######################################################
; now do the fit ...
;#######################################################

   patchnr=1L
   FOR ifr=nfr,nmin_frq,-1 DO BEGIN
      print,'Calculating RM from ',ifr,' frequencies ...'
      nitab=lonarr(ifr)
      FOR i=0,ifr-1 DO BEGIN
         nitab(i)=nfr-ifr+i
      END
      iend=0
      WHILE (iend EQ 0) DO BEGIN
; Prepare mask ...
         print,'Combination: ',nitab
         smask_part(*,*)=1.
         found=0
         mygroup=0
         IF TOTAL(frq_need(nitab)) GE frq_need_sum THEN BEGIN
            FOR i=0,ifr-1 DO BEGIN
               smask_tmp(*,*)=smask_stack(nitab(i),*,*)
               jj=WHERE((smask GT 0.) AND (smask_tmp GT 0.) AND (smask_part GT 0.))
               IF jj(0) NE -1 THEN BEGIN
                  found=1
                  smask_part(*,*)=0.
                  smask_part(jj)=1.
; check that frequencies of different groups are present in the nitab subset -> mgroup < 0
                  IF mygroup EQ 0 THEN BEGIN
                     mygroup=group(nitab(i))
                  END ELSE BEGIN
                     IF ((mygroup GT 0) AND (mygroup NE group(nitab(i)))) THEN mygroup=-group(nitab(i))
                  END
               END
            END
         END ELSE BEGIN
            print,'   Skipped due to frq_need condition ...'
         END
         IF mygroup GE 0 THEN print,'   Skipped due to group condition ...'
         IF (found GT 0) AND (mygroup LT 0) THEN BEGIN
            found=0
            frq_part=frq(nitab)
            lamb2=(2.99792458d+8/(frq_part*1.0d6))^2
            FOR ix=0,dx-1 DO BEGIN
               FOR iy=0,dy-1 DO BEGIN
                  IF smask_part(ix,iy) GT 0. THEN BEGIN
                     S_BAR=0.0d0
                     L2_BAR=0.0d0
                     L4_BAR=0.0d0
                     FOR f=0,ifr-1 DO BEGIN
                        sig2 = (serr_stack(nitab(f),ix,iy)/180.0*!Pi)^2
                        S_BAR = S_BAR + 1./sig2
                        L2_BAR = L2_BAR + lamb2(f)/sig2
                        L4_BAR = L4_BAR + lamb2(f)^2/sig2
                     END
                     squality(ix,iy)=SQRT(S_BAR/(S_BAR*L4_BAR - L2_BAR^2))
                     nrmask(ix,iy)=ifr
;                     smask(ix,iy)=0.0
                     found=1
                  END
               END
            END
;            jjqual=WHERE((squality GT 0.0) AND (nrmask EQ ifr))
            jjqual=WHERE((squality GT 0.0) AND (smask_part GT 0.))
             start_qual=start_tresh(nfr-ifr)
;            IF IS_DEF(start_abs) THEN BEGIN
;               start_qual=start_tresh(nfr-ifr)/sqrt(ifr)
;            END ELSE BEGIN
;               start_qual=MIN(squality(jjqual))*start_tresh(nfr-ifr)
;            END
; prepare arrays
            frq_part=frq(nitab)
            smap_stack_part=smap_stack(nitab,*,*)
            sphia_stack_part=sphia_stack(nitab,*,*)
            serr_stack_part=serr_stack(nitab,*,*)
;            print,min(smap_stack_part),max(smap_stack_part),min(serr_stack_part),max(serr_stack_part)
;            stop
            squality_part=squality(*,*)
            patch_map_part=patch_map(*,*)

; Do some debug
;            jjt=WHERE(smask_part GT 0.)
;            sst=SIZE(jjt)
;            namett='test'
;            FOR it=0,ifr-1 DO namett=namett+string(nitab(it),FORMAT='(i1)')
;            namett=namett+'.txt'
;            print,'Saving in ',namett
;            openw,13,namett
;            FOR jt=0,ifr-1 DO BEGIN
;               tmpmp1=smap_stack_part(jt,*,*)
;               tmpmp2=serr_stack_part(jt,*,*)
;               printf,13,'Maps ',jt
;               FOR it=0L,sst(1)-1 DO BEGIN
;                  printf,13,tmpmp1(jjt(it)),tmpmp2(jjt(it))
;               END
;            END
;            close,13
; Do some debug

; Calculate ...
            make_smooth,smap_stack_part,sphia_stack_part,serr_stack_part,smask_part,frq_part,$
                        rmout,squality_part,$
                        rmout_part,iphi_part,xi2map_part,rejects,$
                        rm1map_part,phi1map_part,flags1_part,patch_map_part,$
                        start_qual=start_qual,do_rm0=do_rm0,$
                        show=show,zoom=zoom,noweight=noweight,rmmin=rmmin,$
                        warning=warning,grad_tol=grad_tol,rmmax=rmmax,save=save,$
                        max_dphi_toll=max_dphi_toll(nfr-ifr),democratic=democratic,$
                        aips=aips,xi2_frac_cut=xi2_frac_cut,min_neighb=min_neighb(nfr-ifr)

            FOR i=0,ifr-1 DO BEGIN
               smap_stack(nitab(i),*,*)=smap_stack_part(i,*,*)
               sphia_stack(nitab(i),*,*)=sphia_stack_part(i,*,*)
            END
            jj=WHERE(rmout_part NE 0.0)
            IF jj(0) GT -1 THEN BEGIN
               rmout(jj)=rmout_part(jj)
               iphi(jj)=iphi_part(jj)
               xi2map(jj)=xi2map_part(jj)
               phi1map(jj)=phi1map_part(jj)
               flags1(jj)=flags1_part(jj)
               patch_map(jj)=patch_map_part(jj)
               smask(jj)=0.0
               jj=WHERE(rm1map_part NE 0.0)
               IF(jj(0)) GT -1 THEN rm1map(jj)=rm1map_part(jj)
; Afterburn
               IF IS_DEF(afterburn) AND (ifr LT nfr) THEN BEGIN
                  print,'Turning on Afterburner ...'
                  jj=WHERE(rmout_part NE 0.0)
                  ssa=SIZE(jj)
; Constructing access map ...
                  ssp=SIZE(rmout_part)
                  xsize=ssp(1)
                  ysize=ssp(2)
                  ixmap=fltarr(xsize,ysize)
                  iymap=fltarr(xsize,ysize)
                  FOR ix=0,xsize-1 DO BEGIN
                     FOR iy=0,ysize-1 DO BEGIN
                        ixmap(ix,iy)=ix
                        iymap(ix,iy)=iy
                     END
                  END
; Go over hole patch
                  FOR i=0,ssa(1)-1 DO BEGIN
                     smap_stack_part=fltarr(nfr)
                     serr_stack_part=fltarr(nfr)
                     frq_part=fltarr(nfr)
                     ix=ixmap(jj(i))
                     iy=iymap(jj(i))
; Copy used values
                     FOR j=0,ifr-1 DO BEGIN
                        smap_stack_part(j)=smap_stack(nitab(j),ix,iy)
                        serr_stack_part(j)=serr_stack(nitab(j),ix,iy)
                        frq_part(j)=frq(nitab(j))
                     END
; Calculate remaining frq's
                     rest=indgen(nfr)
                     rest(nitab)=10000
                     ii=SORT(rest)
                     rest=rest(ii(0:nfr-ifr-1))

; Find more frequencies to use
                     tnfr=ifr
                     FOR j=0,nfr-ifr-1 DO BEGIN
                        phi_err_test=serr_stack(rest(j),ix,iy)
                        IF phi_err_test LT phi_err_tre_after(rest(j)) THEN BEGIN
                           smap_stack_part(tnfr)=smap_stack(rest(j),ix,iy)
                           serr_stack_part(tnfr)=phi_err_test
                           frq_part(tnfr)=frq(rest(j))
                           tnfr=tnfr+1
                        END
                     END

; New fit using more frequencies
                     IF tnfr GT ifr THEN BEGIN
                        smap_stack_part=smap_stack_part(0:tnfr-1)
                        serr_stack_part=serr_stack_part(0:tnfr-1)
                        frq_part=frq_part(0:tnfr-1)
                        lamb2=(2.99792458d+8/(frq_part*1.0d6))^2
                        fit_rm,lamb2,smap_stack_part,serr_stack_part,phi_ini_aft,rm_aft,xi2_aft,flag,$
                               rmmax=rmmax,rmmin=rmmin,noweight=noweight,warning=warning,$
                               minf=ifr
                        IF flag(0) GT xi2_frac_cut THEN BEGIN
                           IF afterburn GT 0 THEN BEGIN 
                              print,'Rejecting pixel from afterburner due to marginal best fit n-pi ...',$
                                    flag(0),xi2_frac_cut
                              rmout(jj(i))=0.0
                              nrmask(jj(i))=0
                              patch_map(jj(i))=0
                              xi2map(jj(i))=0
                           END
                        END ELSE BEGIN
;                        print,'Improved RM from ',rmout(jj(i)),' to ',rm_aft(0)
;                         if rm_aft(0)-rmout(jj(i)) LT 1000. THEN stop
; Saving new values
                           flags1(jj(i))=flag(0)
                           rmout(jj(i))=rm_aft(0)
                           iphi(jj(i))=phi_ini_aft(0)
                           xi2map(jj(i))=xi2_aft(0)
                           FOR j=0,tnfr-ifr-1 DO BEGIN
                              smap_stack(rest(j),ix,iy)=smap_stack_part(ifr+j)
                           END
                           nrmask(ix,iy)=tnfr
; Calculating new quality of pixel
                           S_BAR=0.0d0
                           L2_BAR=0.0d0
                           L4_BAR=0.0d0
                           FOR f=0,tnfr-1 DO BEGIN
                              sig2 = (serr_stack_part(f)/180.0*!Pi)^2
                              S_BAR = S_BAR + 1./sig2
                              L2_BAR = L2_BAR + lamb2(f)/sig2
                              L4_BAR = L4_BAR + lamb2(f)^2/sig2
                           END
                           squality(ix,iy)=SQRT(S_BAR/(S_BAR*L4_BAR - L2_BAR^2))
                        END
                     END
;                     stop
                  END
               END
            END
         END
; End Afterburn

; Choose next combination of frq's
         nitab(0)=nitab(0)-1
         FOR i=0,ifr-2 DO BEGIN
            IF nitab(i) LT i THEN BEGIN
               nitab(i+1)=nitab(i+1)-1
               FOR j=0,i DO BEGIN
                  jm=i-j
                  nitab(jm)=nitab(jm+1)-1
               END
            END
         END
         IF (nitab(0) LT 0) THEN iend=1
      END
   END

   rmmap=rmout
   iphimap=iphi

   ssm=size(mask)

   IF IS_DEF(save) THEN BEGIN
; Final RM Map
      print,'Saving RM to ',save
      savemap=fltarr(ssm(1),ssm(2))
      savemap(minix:maxix,miniy:maxiy)=rmmap(*,*)
      jjs=where(savemap EQ 0.0)
      savemap(jjs)=0.0/0.0
      writefits,save+'.fits',savemap,header

;      jj=where((ABS(rmmap) LT 500) AND (rmmap GT 0.0))
;      ss=SIZE(jj)
;      mean=TOTAL(rmmap(jj))/ss(1)
;      rmsnew=SQRT(TOTAL((rmmap(jj)-mean)^2)/ss(1))
;      print,'New RM-fit: npix=',ss(1),' RM=',max(abs(rmmap(jj))),mean,rmsnew

; Xi2 Map
      savemap(*,*)=-1
      savemap(minix:maxix,miniy:maxiy)=xi2map(*,*)
      writefits,save+'_xi2.fits',savemap,header

; Nfrq Map
      savemap(*,*)=-1
      savemap(minix:maxix,miniy:maxiy)=nrmask(*,*)
      writefits,save+'_nfrq.fits',savemap,header

; Rejects Map
      savemap(*,*)=0
      savemap(minix:maxix,miniy:maxiy)=rejects(*,*)
      writefits,save+'_rej.fits',savemap,header

; Phi_0 Map
      savemap(*,*)=0.0
      savemap(minix:maxix,miniy:maxiy)=iphi(*,*)
      jjs=where(savemap EQ 0.0)
      savemap(jjs)=0.0/0.0
      writefits,save+'_phi0.fits',savemap,header

; Quality Map
      savemap(*,*)=0.0
      savemap(minix:maxix,miniy:maxiy)=squality(*,*)
      writefits,save+'_quality.fits',savemap,header

; RM1 Map
      IF IS_DEF(do_rm0) THEN BEGIN
         savemap(*,*)=0.0
         savemap(minix:maxix,miniy:maxiy)=rm1map(*,*)
         jjs=where(savemap EQ 0.0)
         savemap(jjs)=0.0/0.0
         writefits,save+'_rm1.fits',savemap,header

         savemap(*,*)=0.0
         savemap(minix:maxix,miniy:maxiy)=phi1map(*,*)
         jjs=where(savemap EQ 0.0)
         savemap(jjs)=0.0/0.0
         writefits,save+'_rm1_phi0.fits',savemap,header
      END

; flags Map
      savemap(*,*)=0.0
      savemap(minix:maxix,miniy:maxiy)=flags1(*,*)
      writefits,save+'_flags1.fits',savemap,header

; patch Map
      savemap(*,*)=0.0
      savemap(minix:maxix,miniy:maxiy)=patch_map(*,*)
      jjs=where(savemap EQ 0.0)
      IF jjs(0) GT -1 THEN savemap(jjs)=0.0/0.0
      writefits,save+'_patch.fits',savemap,header

; Abs Phi Maps
      FOR ifr=0,nfr-1 DO BEGIN
         savemap(*,*)=0.0
         savemap(minix:maxix,miniy:maxiy)=smap_stack(ifr,*,*)
         jjs=where(savemap EQ 0.0)
         IF jjs(0) GT -1 THEN savemap(jjs)=0.0/0.0
         writefits,save+'_phi'+string(ifr+1,FORMAT='(i1)')+'.fits',savemap,header
      END

   END

   IF IS_DEF(sto) THEN stop

END

